import pytest

from si_tests import settings
from si_tests import logger
from si_tests.clients.iam import keycloak_client
from si_tests.utils import utils
import si_tests.utils.waiters as helpers

LOG = logger.logger

FIRST_UUID_PART_LENGTH = 8


@pytest.fixture
def keycloak_client_fixture(kaas_manager):
    keycloak_ip = kaas_manager.get_keycloak_ip()
    pwd = kaas_manager.get_secret_data(
        "iam-api-secrets", "kaas", "keycloak_password")
    kclient = keycloak_client.KeycloakAdminClient(
        ip=keycloak_ip, user=settings.KEYCLOAK_USER, password=pwd)
    return kclient


@pytest.fixture
def keycloak_user_fixture(kaas_manager, keycloak_client_fixture):
    username = utils.gen_random_string(6)
    password = utils.gen_random_string(10)
    uid = keycloak_client_fixture.user_create(username=username,
                                              password=password)
    iamusername = f"{username}-{uid[:FIRST_UUID_PART_LENGTH]}"

    def user_created():
        user = None
        try:
            user = kaas_manager.get_iamuser(iamusername).read().to_dict()
        except Exception:
            pass

        return user is not None

    helpers.wait(
        user_created, timeout=150, interval=15,
        timeout_msg="Timeout waiting for iamuser.")

    yield username, password, uid, iamusername

    keycloak_client_fixture.user_delete(username)

    def user_deleted():
        user_found = True
        try:
            kaas_manager.get_iamuser(iamusername).read()
        except Exception as e:
            user_found = e.status != 404
        return not user_found

    helpers.wait(
        user_deleted, timeout=150, interval=15,
        timeout_msg="Timeout cleaning up iamuser.")


def find_user_realm_roles(kclient, userid, target_roles):
    roles = kclient.get_user_roles(userid)
    LOG.info("User %s roles are: %s; target roles are %s",
             userid, roles, target_roles)

    roles_copy = target_roles[:]
    for role in roles:
        if role["name"] in roles_copy:
            roles_copy.remove(role["name"])
    return len(roles_copy) == 0


def test_add_iamrolebinding_user_failed(kaas_manager):
    """
    1. Apply IAMRoleBinding with incorrect username.
    """
    rnd_string = utils.gen_random_string(6)
    ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)

    iamrolebinding_name = "{rnd_string}-iamrb".format(rnd_string=rnd_string)
    user_name = "{rnd_string}-user".format(rnd_string=rnd_string)
    LOG.info("IAMRoleBinding name - {iamrolebinding_name}, "
             "username - {username}".format(
                 iamrolebinding_name=iamrolebinding_name,
                 username=user_name)
             )
    try:
        ns.create_iamrolebinding(iamrolebinding_name, "operator", user_name)
    except Exception as e:
        assert e.status == 400, ('Wrong response code. {}'.format(e))
    else:
        pytest.fail("It is possible to create iamrolebinding"
                    "with wrong username")


def test_add_iamrolebinding_role_failed(kaas_manager, keycloak_user_fixture):
    """
    2. Create a new user.
    3. Apply IAMRoleBinding with incorrect rolename.
    4. Clean up user.
    """
    rnd_string = utils.gen_random_string(6)
    ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)
    _, __, ___, iamusername = keycloak_user_fixture

    iamrolebinding_name = "{rnd_string}-iamrb".format(rnd_string=rnd_string)
    role_name = "{rnd_string}-role".format(rnd_string=rnd_string)
    LOG.info("IAMRoleBinding name - {iamrolebinding_name}, "
             "rolename - {rolename}".format(
                 iamrolebinding_name=iamrolebinding_name,
                 rolename=role_name)
             )
    try:
        ns.create_iamrolebinding(iamrolebinding_name, role_name,
                                 iamusername)
    except Exception as e:
        assert e.status == 400, ('Wrong response code. {}'.format(e))
    else:
        pytest.fail("It is possible to create iamrolebinding"
                    "with wrong rolename")


def test_add_iamclusterrolebinding_cluster_failed(
        kaas_manager, keycloak_user_fixture):
    """
    5. Create a new user.
    6. Apply IAMClusterRoleBinding with incorrect clustername.
    7. Clean up user.
    """
    rnd_string = utils.gen_random_string(6)
    ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)
    _, __, ___, iamusername = keycloak_user_fixture

    iamrolebinding_name = "{rnd_string}-iamrb".format(rnd_string=rnd_string)
    cluster_name = "{rnd_string}-cluster".format(rnd_string=rnd_string)
    LOG.info("IAMClusterRoleBinding name - {iamrolebinding_name}, "
             "clustername - {clustername}".format(
                 iamrolebinding_name=iamrolebinding_name,
                 clustername=cluster_name)
             )
    try:
        ns.create_iamclusterrolebinding(
            iamrolebinding_name, "operator", iamusername, cluster_name)
    except Exception as e:
        assert e.status == 400, ('Wrong response code. {}'.format(e))
    else:
        pytest.fail("It is possible to create iamrolebinding"
                    "with wrong cluster name")


def test_add_iamrolebinding_success(
        kaas_manager, keycloak_client_fixture, keycloak_user_fixture):
    """
    8. Create a new user.
    9. Check that user with no roles can not access iamrolebindings.
    10. Create IAMRoleBinding with correct parameters.
    11. Request roles list from KeyCloak to check role presence.
    12. Check that user with operator role can access iamrolebindings.
    13. Remove IAMRoleBinding from k8s API.
    14. Request roles list from KeyCloak to check role absence.
    15. Create another IAMRoleBinding.
    16. Clean up user. Check that IAMUser object is cleaned up even if
        IAMRoleBinding is present before deletion.
    """
    username, password, uid, iamusername = keycloak_user_fixture

    check_resource_path = (
        f"apis/iam.mirantis.com/v1alpha1/namespaces/"
        f"{settings.TARGET_NAMESPACE}/iamrolebindings")
    ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)
    cluster = ns.get_cluster(settings.TARGET_CLUSTER)
    # Negative check
    LOG.info(f"Login to UI using {username}, and check that "
             f"access to the {check_resource_path} is restricted")
    d_client = cluster.get_mcc_dashboardclient(
        user=username, password=password)
    d_resources = d_client.get_resource(
        check_resource_path, raise_on_error=False)
    LOG.debug(f"Response from {check_resource_path} using a new user "
              f"(must be restricted for the user with no roles):\n"
              f"{d_resources}")
    assert 'code' in d_resources and d_resources['code'] == 403, (
        f"Access to {check_resource_path} is allowed "
        f"for the user with no roles, while it should be restricted")

    rnd_string = utils.gen_random_string(6)
    iamrolebinding_name = "{rnd_string}-iamrb".format(rnd_string=rnd_string)
    ns.create_iamrolebinding(iamrolebinding_name, "operator", iamusername)
    target_roles = ["m:kaas:%s@operator" % settings.TARGET_NAMESPACE]
    helpers.wait(
        lambda: find_user_realm_roles(
            keycloak_client_fixture, uid, target_roles),
        timeout=15, interval=5,
        timeout_msg="Timeout waiting for keycloak role assignment.")

    # Re-login with new grants
    LOG.info(f"Re-login to UI using {username}, and check"
             f"that access to the {check_resource_path} is allowed now")
    d_client = cluster.get_mcc_dashboardclient(
        user=username, password=password)
    # Positive check
    d_resources = d_client.get_resource(check_resource_path)
    LOG.debug(f"{check_resource_path} read:\n{d_resources}")

    iamrb = ns.get_iamrolebinding(iamrolebinding_name)
    iamrb.delete()
    helpers.wait(
        lambda: not find_user_realm_roles(
            keycloak_client_fixture, uid, target_roles),
        timeout=15, interval=5,
        timeout_msg="Timeout waiting for keycloak role removal.")

    rnd_string = utils.gen_random_string(6)
    iamrolebinding_name = "{rnd_string}-iamrb".format(rnd_string=rnd_string)
    ns.create_iamrolebinding(iamrolebinding_name, "user", iamusername)
    target_roles = ["m:kaas:%s@user" % settings.TARGET_NAMESPACE]
    helpers.wait(
        lambda: find_user_realm_roles(
            keycloak_client_fixture, uid, target_roles),
        timeout=15, interval=5,
        timeout_msg="Timeout waiting for keycloak role assignment.")


def test_add_iamuser_keycloak_external_legacy_reader(
        kaas_manager, keycloak_client_fixture, keycloak_user_fixture):
    """
    17. Create a user.
    18. Check that user with no roles can not list namespaces.
    19. Assign legacy m:kaas@reader role in keycloak.
    20. Check if there is a proper IAMGlobalRoleBinding object.
    21. Check that user with can list namespaces now.
    22. Set IAMGlobalRoleBinding legacy flag to false.
    23. Check if legacy roles' style changed to new in keycloak
    24. Create new namespace.
    25. Check that role corresponding to new namespace is assigned to user.
    26. Delete namespace.
    27. Check that corresponding role is deleted.
    28. Check that user still can list namespaces.
    29. Cleanup user.
    """
    username, password, uid, iamusername = keycloak_user_fixture

    check_resource_path = "api/v1/namespaces"
    ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)
    cluster = ns.get_cluster(settings.TARGET_CLUSTER)
    # Negative check
    LOG.info(f"Login to UI using {username}, and check that "
             f"access to the {check_resource_path} is restricted")
    d_client = cluster.get_mcc_dashboardclient(
        user=username, password=password)
    d_resources = d_client.get_resource(
        check_resource_path, raise_on_error=False)
    LOG.debug(f"Response from {check_resource_path} using a new user "
              f"(must be restricted for the user with no roles):\n"
              f"{d_resources}")
    assert 'code' in d_resources and d_resources['code'] == 403, (
        f"Access to {check_resource_path} is allowed "
        f"for the user with no roles, while it should be restricted")

    scope = "m:kaas"
    role = "reader"
    roles = ["{scope}@{role}".format(scope=scope, role=role)]
    keycloak_client_fixture.assign_roles(username, roles=roles)

    expected_rb_name = "{username}-{role}".format(
        username=iamusername, role="user")

    def globalrolebinding_created():
        globalrb = None
        try:
            globalrb = kaas_manager.get_iamglobalrolebinding(
                expected_rb_name).read().to_dict()
        except Exception:
            pass

        return globalrb is not None

    helpers.wait(
        globalrolebinding_created, timeout=150, interval=15,
        timeout_msg="Timeout waiting for iamglobalrolebinding.")

    # Re-login with new grants
    LOG.info(f"Re-login to UI using {username}, and check"
             f"that access to the {check_resource_path} is allowed now")
    d_client = cluster.get_mcc_dashboardclient(
        user=username, password=password)
    # Positive check
    d_resources = d_client.get_resource(check_resource_path)
    LOG.debug(f"{check_resource_path} read:\n{d_resources}")

    globalrb = kaas_manager.get_iamglobalrolebinding(expected_rb_name)
    globalrb.patch({"legacy": False})

    target_namespaces = [
        ns.name for ns in kaas_manager.get_namespaces()] + ["default"]
    target_roles = ["m:kaas:%s@user" % ns for ns in target_namespaces]

    helpers.wait(
        lambda: find_user_realm_roles(
            keycloak_client_fixture, uid, target_roles),
        timeout=15, interval=5,
        timeout_msg="Timeout waiting for keycloak role change.")

    ns = kaas_manager.create_namespace(username)
    helpers.wait(
        lambda: find_user_realm_roles(
            keycloak_client_fixture, uid, ["m:kaas:%s@user" % username]),
        timeout=60, interval=5,
        timeout_msg="Timeout waiting for keycloak role addition.")
    ns.delete()
    helpers.wait(
        lambda: not find_user_realm_roles(
            keycloak_client_fixture, uid, ["m:kaas:%s@user" % username]),
        timeout=60, interval=5,
        timeout_msg="Timeout waiting for keycloak role removal.")

    # New style roles should work the same way
    LOG.info(f"Re-login to UI using {username}, and check"
             f"that access to the {check_resource_path} is allowed")
    d_client = cluster.get_mcc_dashboardclient(
        user=username, password=password)
    # Positive check
    d_resources = d_client.get_resource(check_resource_path)
    LOG.debug(f"{check_resource_path} read:\n{d_resources}")


def test_add_iamuser_keycloak_external_legacy_writer(
        kaas_manager, keycloak_client_fixture, keycloak_user_fixture):
    """
    30. Create a user.
    31. Check that user with no roles can not list iamglobalrolebindings.
    32. Assign legacy m:kaas@writer role in keycloak.
    33. Check if there are corresponding IAMGlobalRoleBinding objects.
    34. Check that user with can list iamglobalrolebindings now.
    35. Set IAMGlobalRoleBindings' legacy flag to false.
    36. Check if legacy roles' style changed to new in keycloak
    37. Check that user still can list iamglobalrolebindings.
    38. Cleanup user.
    """
    username, password, uid, iamusername = keycloak_user_fixture

    check_resource_path = (
        "apis/iam.mirantis.com/v1alpha1/iamglobalrolebindings")
    ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)
    cluster = ns.get_cluster(settings.TARGET_CLUSTER)
    # Negative check
    LOG.info(f"Login to UI using {username}, and check that "
             f"access to the {check_resource_path} is restricted")
    d_client = cluster.get_mcc_dashboardclient(
        user=username, password=password)
    d_resources = d_client.get_resource(
        check_resource_path, raise_on_error=False)
    LOG.debug(f"Response from {check_resource_path} using a new user "
              f"(must be restricted for the user with no roles):\n"
              f"{d_resources}")
    assert 'code' in d_resources and d_resources['code'] == 403, (
        f"Access to {check_resource_path} is allowed "
        f"for the user with no roles, while it should be restricted")

    scope = "m:kaas"
    role = "writer"
    roles = ["{scope}@{role}".format(scope=scope, role=role)]
    keycloak_client_fixture.assign_roles(username, roles=roles)

    expected_rb_names = [
        "{username}-{role}".format(username=iamusername, role=r)
        for r in ("operator", "global-admin")]

    def globalrolebindings_created():
        globalrbs = []
        for expected_rb_name in expected_rb_names:
            try:
                globalrbs.append(
                    kaas_manager.get_iamglobalrolebinding(
                        expected_rb_name).read().to_dict())
            except Exception:
                pass

        return len(globalrbs) == 2

    helpers.wait(
        globalrolebindings_created, timeout=150, interval=15,
        timeout_msg="Timeout waiting for iamglobalrolebindings.")

    # Re-login with new grants
    LOG.info(f"Re-login to UI using {username}, and check"
             f"that access to the {check_resource_path} is allowed now")
    d_client = cluster.get_mcc_dashboardclient(
        user=username, password=password)
    # Positive check
    d_resources = d_client.get_resource(check_resource_path)
    LOG.debug(f"{check_resource_path} read:\n{d_resources}")

    for expected_rb_name in expected_rb_names:
        globalrb = kaas_manager.get_iamglobalrolebinding(expected_rb_name)
        globalrb.patch({"legacy": False})

    target_namespaces = [
        ns.name for ns in kaas_manager.get_namespaces()] + ["default"]
    target_roles = ["m:kaas:%s@operator" % ns for ns in target_namespaces]
    target_roles.append("m:kaas@global-admin")

    helpers.wait(
        lambda: find_user_realm_roles(
            keycloak_client_fixture, uid, target_roles),
        timeout=15, interval=5,
        timeout_msg="Timeout waiting for keycloak role change.")
    helpers.wait(
        lambda: not find_user_realm_roles(
            keycloak_client_fixture, uid, ["m:kaas@writer"]),
        timeout=60, interval=5,
        timeout_msg="Timeout waiting for keycloak role removal.")

    # New style roles should work the same way
    LOG.info(f"Re-login to UI using {username}, and check"
             f"that access to the {check_resource_path} is allowed")
    d_client = cluster.get_mcc_dashboardclient(
        user=username, password=password)
    # Positive check
    d_resources = d_client.get_resource(check_resource_path)
    LOG.debug(f"{check_resource_path} read:\n{d_resources}")
